# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements TensorFlow quaternion utility functions.

A quaternion is written as $$q =  xi + yj + zk + w$$, where $$i,j,k$$ forms the
three bases of the imaginary part. The functions implemented in this file
use the Hamilton convention where $$i^2 = j^2 = k^2 = ijk = -1$$. A quaternion
is stored in a 4-D vector $$[x, y, z, w]^T$$.

More details about Hamiltonian quaternions can be found on [this page.]
(https://en.wikipedia.org/wiki/Quaternion)

Note: Some of the functions expect normalized quaternions as inputs where
$$x^2 + y^2 + z^2 + w^2 = 1$$.

"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf

from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.math import vector
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import safe_ops
from tensorflow_graphics.util import shape


def _build_quaternion_from_sines_and_cosines(sin_half_angles, cos_half_angles):
  """Builds a quaternion from sines and cosines of half Euler angles.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    sin_half_angles: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the sine of half Euler angles.
    cos_half_angles: A tensor of shape `[A1, ..., An, 3]`, where the last
      dimension represents the cosine of half Euler angles.

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a quaternion.
  """
  c1, c2, c3 = tf.unstack(cos_half_angles, axis=-1)
  s1, s2, s3 = tf.unstack(sin_half_angles, axis=-1)
  w = c1 * c2 * c3 + s1 * s2 * s3
  x = -c1 * s2 * s3 + s1 * c2 * c3
  y = c1 * s2 * c3 + s1 * c2 * s3
  z = -s1 * s2 * c3 + c1 * c2 * s3
  return tf.stack((x, y, z, w), axis=-1)


def between_two_vectors_3d(vector1, vector2, name=None):
  """Computes quaternion over the shortest arc between two vectors.

  Result quaternion describes shortest geodesic rotation from
  vector1 to vector2.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector1: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the first vector.
    vector2: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the second vector.
    name: A name for this op that defaults to
      "quaternion_between_two_vectors_3d".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `vector1` or `vector2` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_between_two_vectors_3d",
                               [vector1, vector2]):
    vector1 = tf.convert_to_tensor(value=vector1)
    vector2 = tf.convert_to_tensor(value=vector2)

    shape.check_static(
        tensor=vector1, tensor_name="vector1", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=vector2, tensor_name="vector2", has_dim_equals=(-1, 3))
    shape.compare_batch_dimensions(
        tensors=(vector1, vector2), last_axes=-2, broadcast_compatible=True)

    # Make sure that we are dealing with unit vectors.
    vector1 = tf.nn.l2_normalize(vector1, axis=-1)
    vector2 = tf.nn.l2_normalize(vector2, axis=-1)
    cos_theta = vector.dot(vector1, vector2)
    real_part = 1.0 + cos_theta
    axis = vector.cross(vector1, vector2)

    # Compute arbitrary antiparallel axes to rotate around in case of opposite
    # vectors.
    x, y, z = tf.split(vector1, (1, 1, 1), axis=-1)
    x_bigger_z = tf.abs(x) > tf.abs(z)
    x_bigger_z = tf.concat([x_bigger_z] * 3, axis=-1)
    antiparallel_axis = tf.compat.v1.where(
        x_bigger_z, tf.concat((-y, x, tf.zeros_like(z)), axis=-1),
        tf.concat((tf.zeros_like(x), -z, y), axis=-1))

    # Compute rotation between two vectors.
    is_antiparallel = real_part < 1e-6
    is_antiparallel = tf.concat([is_antiparallel] * 4, axis=-1)
    rot = tf.compat.v1.where(
        is_antiparallel,
        tf.concat((antiparallel_axis, tf.zeros_like(real_part)), axis=-1),
        tf.concat((axis, real_part), axis=-1))
    return tf.nn.l2_normalize(rot, axis=-1)


def conjugate(quaternion, name=None):
  """Computes the conjugate of a quaternion.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion: A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    name: A name for this op that defaults to "quaternion_conjugate".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_conjugate", [quaternion]):
    quaternion = tf.convert_to_tensor(value=quaternion)

    shape.check_static(
        tensor=quaternion, tensor_name="quaternion", has_dim_equals=(-1, 4))
    quaternion = asserts.assert_normalized(quaternion)

    xyz, w = tf.split(quaternion, (3, 1), axis=-1)
    return tf.concat((-xyz, w), axis=-1)


def from_axis_angle(axis, angle, name=None):
  """Converts an axis-angle representation to a quaternion.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    axis: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a normalized axis.
    angle: A tensor of shape `[A1, ..., An, 1]`, where the last dimension
      represents an angle.
    name: A name for this op that defaults to "quaternion_from_axis_angle".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `axis` or `angle` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_from_axis_angle",
                               [axis, angle]):
    axis = tf.convert_to_tensor(value=axis)
    angle = tf.convert_to_tensor(value=angle)

    shape.check_static(tensor=axis, tensor_name="axis", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=angle, tensor_name="angle", has_dim_equals=(-1, 1))
    shape.compare_batch_dimensions(
        tensors=(axis, angle), last_axes=-2, broadcast_compatible=True)
    axis = asserts.assert_normalized(axis)

    half_angle = 0.5 * angle
    w = tf.cos(half_angle)
    xyz = tf.sin(half_angle) * axis
    return tf.concat((xyz, w), axis=-1)


def from_euler(angles, name=None):
  """Converts an Euler angle representation to a quaternion.

  Note:
    Uses the z-y-x rotation convention (Tait-Bryan angles).

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    angles: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents the three Euler angles. `[..., 0]` is the angle about `x` in
      radians, `[..., 1]` is the angle about `y` in radians and `[..., 2]` is
      the angle about `z` in radians.
    name: A name for this op that defaults to "quaternion_from_euler".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `angles` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_from_euler", [angles]):
    angles = tf.convert_to_tensor(value=angles)

    shape.check_static(
        tensor=angles, tensor_name="angles", has_dim_equals=(-1, 3))

    half_angles = angles / 2.0
    cos_half_angles = tf.cos(half_angles)
    sin_half_angles = tf.sin(half_angles)
    return _build_quaternion_from_sines_and_cosines(sin_half_angles,
                                                    cos_half_angles)


def from_euler_with_small_angles_approximation(angles, name=None):
  r"""Converts small Euler angles to quaternions.

  Under the small angle assumption, $$\sin(x)$$ and $$\cos(x)$$ can be
  approximated by their second order Taylor expansions, where
  $$\sin(x) \approx x$$ and $$\cos(x) \approx 1 - \frac{x^2}{2}$$.
  In the current implementation, the smallness of the angles is not verified.

  Note:
    Uses the z-y-x rotation convention (Tait-Bryan angles).

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
   angles: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
     represents the three Euler angles. `[..., 0]` is the angle about `x` in
     radians, `[..., 1]` is the angle about `y` in radians and `[..., 2]` is the
     angle about `z` in radians.
    name: A name for this op that defaults to "quaternion_from_euler".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `angles` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_from_euler", [angles]):
    angles = tf.convert_to_tensor(value=angles)

    shape.check_static(
        tensor=angles, tensor_name="angles", has_dim_equals=(-1, 3))

    half_angles = angles / 2.0
    cos_half_angles = 1.0 - 0.5 * half_angles * half_angles
    sin_half_angles = half_angles
    quaternion = _build_quaternion_from_sines_and_cosines(
        sin_half_angles, cos_half_angles)
    # We need to normalize the quaternion due to the small angle approximation.
    return tf.nn.l2_normalize(quaternion, axis=-1)


def from_rotation_matrix(rotation_matrix, name=None):
  """Converts a rotation matrix representation to a quaternion.

  Warning:
    This function is not smooth everywhere.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    rotation_matrix: A tensor of shape `[A1, ..., An, 3, 3]`, where the last two
      dimensions represent a rotation matrix.
    name: A name for this op that defaults to "quaternion_from_rotation_matrix".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `rotation_matrix` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_from_rotation_matrix",
                               [rotation_matrix]):
    rotation_matrix = tf.convert_to_tensor(value=rotation_matrix)

    shape.check_static(
        tensor=rotation_matrix,
        tensor_name="rotation_matrix",
        has_rank_greater_than=1,
        has_dim_equals=((-1, 3), (-2, 3)))
    rotation_matrix = rotation_matrix_3d.assert_rotation_matrix_normalized(
        rotation_matrix)

    trace = tf.linalg.trace(rotation_matrix)
    eps_addition = asserts.select_eps_for_addition(rotation_matrix.dtype)
    rows = tf.unstack(rotation_matrix, axis=-2)
    entries = [tf.unstack(row, axis=-1) for row in rows]

    def tr_positive():
      sq = tf.sqrt(trace + 1.0) * 2.  # sq = 4 * qw.
      qw = 0.25 * sq
      qx = safe_ops.safe_unsigned_div(entries[2][1] - entries[1][2], sq)
      qy = safe_ops.safe_unsigned_div(entries[0][2] - entries[2][0], sq)
      qz = safe_ops.safe_unsigned_div(entries[1][0] - entries[0][1], sq)
      return tf.stack((qx, qy, qz, qw), axis=-1)

    def cond_1():
      sq = tf.sqrt(1.0 + entries[0][0] - entries[1][1] - entries[2][2] +
                   eps_addition) * 2.  # sq = 4 * qx.
      qw = safe_ops.safe_unsigned_div(entries[2][1] - entries[1][2], sq)
      qx = 0.25 * sq
      qy = safe_ops.safe_unsigned_div(entries[0][1] + entries[1][0], sq)
      qz = safe_ops.safe_unsigned_div(entries[0][2] + entries[2][0], sq)
      return tf.stack((qx, qy, qz, qw), axis=-1)

    def cond_2():
      sq = tf.sqrt(1.0 + entries[1][1] - entries[0][0] - entries[2][2] +
                   eps_addition) * 2.  # sq = 4 * qy.
      qw = safe_ops.safe_unsigned_div(entries[0][2] - entries[2][0], sq)
      qx = safe_ops.safe_unsigned_div(entries[0][1] + entries[1][0], sq)
      qy = 0.25 * sq
      qz = safe_ops.safe_unsigned_div(entries[1][2] + entries[2][1], sq)
      return tf.stack((qx, qy, qz, qw), axis=-1)

    def cond_3():
      sq = tf.sqrt(1.0 + entries[2][2] - entries[0][0] - entries[1][1] +
                   eps_addition) * 2.  # sq = 4 * qz.
      qw = safe_ops.safe_unsigned_div(entries[1][0] - entries[0][1], sq)
      qx = safe_ops.safe_unsigned_div(entries[0][2] + entries[2][0], sq)
      qy = safe_ops.safe_unsigned_div(entries[1][2] + entries[2][1], sq)
      qz = 0.25 * sq
      return tf.stack((qx, qy, qz, qw), axis=-1)

    def cond_idx(cond):
      cond = tf.expand_dims(cond, -1)
      cond = tf.tile(cond, [1] * (rotation_matrix.shape.ndims - 2) + [4])
      return cond

    where_2 = tf.compat.v1.where(
        cond_idx(entries[1][1] > entries[2][2]), cond_2(), cond_3())
    where_1 = tf.compat.v1.where(
        cond_idx((entries[0][0] > entries[1][1])
                 & (entries[0][0] > entries[2][2])), cond_1(), where_2)
    quat = tf.compat.v1.where(cond_idx(trace > 0), tr_positive(), where_1)
    return quat


def inverse(quaternion, name=None):
  """Computes the inverse of a quaternion.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion:  A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    name: A name for this op that defaults to "quaternion_inverse".

  Returns:
    A tensor of shape `[A1, ..., An, 4]`, where the last dimension represents
    a normalized quaternion.

  Raises:
    ValueError: If the shape of `quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_inverse", [quaternion]):
    quaternion = tf.convert_to_tensor(value=quaternion)

    shape.check_static(
        tensor=quaternion, tensor_name="quaternion", has_dim_equals=(-1, 4))
    quaternion = asserts.assert_normalized(quaternion)

    squared_norm = tf.reduce_sum(
        input_tensor=tf.square(quaternion), axis=-1, keepdims=True)
    return safe_ops.safe_unsigned_div(conjugate(quaternion), squared_norm)


def is_normalized(quaternion, atol=1e-3, name=None):
  """Determines if quaternion is normalized quaternion or not.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion:  A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a quaternion.
    atol: The absolute tolerance parameter.
    name: A name for this op that defaults to "quaternion_is_normalized".

  Returns:
    A tensor of type `bool` and shape `[A1, ..., An, 1]`, where False indicates
    that the quaternion is not normalized.

  Raises:
    ValueError: If the shape of `quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_is_normalized", [quaternion]):
    quaternion = tf.convert_to_tensor(value=quaternion)

    shape.check_static(
        tensor=quaternion, tensor_name="quaternion", has_dim_equals=(-1, 4))

    norms = tf.norm(tensor=quaternion, axis=-1, keepdims=True)
    return tf.compat.v1.where(
        tf.abs(norms - 1.) < atol, tf.ones_like(norms, dtype=bool),
        tf.zeros_like(norms, dtype=bool))


def normalize(quaternion, eps=1e-12, name=None):
  """Normalizes a quaternion.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion:  A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a quaternion.
    eps: A lower bound value for the norm that defaults to 1e-12.
    name: A name for this op that defaults to "quaternion_normalize".

  Returns:
    A N-D tensor of shape `[?, ..., ?, 1]` where the quaternion elements have
    been normalized.

  Raises:
    ValueError: If the shape of `quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_normalize", [quaternion]):
    quaternion = tf.convert_to_tensor(value=quaternion)

    shape.check_static(
        tensor=quaternion, tensor_name="quaternion", has_dim_equals=(-1, 4))

    return tf.math.l2_normalize(quaternion, axis=-1, epsilon=eps)


def multiply(quaternion1, quaternion2, name=None):
  """Multiplies two quaternions.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion1:  A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a quaternion.
    quaternion2:  A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a quaternion.
    name: A name for this op that defaults to "quaternion_multiply".

  Returns:
    A tensor of shape `[A1, ..., An, 4]` representing quaternions.

  Raises:
    ValueError: If the shape of `quaternion1` or `quaternion2` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_multiply",
                               [quaternion1, quaternion2]):
    quaternion1 = tf.convert_to_tensor(value=quaternion1)
    quaternion2 = tf.convert_to_tensor(value=quaternion2)

    shape.check_static(
        tensor=quaternion1, tensor_name="quaternion1", has_dim_equals=(-1, 4))
    shape.check_static(
        tensor=quaternion2, tensor_name="quaternion2", has_dim_equals=(-1, 4))

    x1, y1, z1, w1 = tf.unstack(quaternion1, axis=-1)
    x2, y2, z2, w2 = tf.unstack(quaternion2, axis=-1)
    x = x1 * w2 + y1 * z2 - z1 * y2 + w1 * x2
    y = -x1 * z2 + y1 * w2 + z1 * x2 + w1 * y2
    z = x1 * y2 - y1 * x2 + z1 * w2 + w1 * z2
    w = -x1 * x2 - y1 * y2 - z1 * z2 + w1 * w2
    return tf.stack((x, y, z, w), axis=-1)


def normalized_random_uniform(quaternion_shape, name=None):
  """Random normalized quaternion following a uniform distribution law on SO(3).

  Args:
    quaternion_shape: A list representing the shape of the output tensor.
    name: A name for this op that defaults to
      "quaternion_normalized_random_uniform".

  Returns:
    A tensor of shape `[quaternion_shape[0],...,quaternion_shape[-1], 4]`
    representing random normalized quaternions.
  """
  with tf.compat.v1.name_scope(name, "quaternion_normalized_random_uniform",
                               [quaternion_shape]):
    quaternion_shape = tf.convert_to_tensor(value=quaternion_shape,
                                            dtype=tf.int32)
    quaternion_shape = tf.concat((quaternion_shape, tf.constant([4])), axis=0)
    random_normal = tf.random.normal(quaternion_shape)
  return normalize(random_normal)


def normalized_random_uniform_initializer():
  """Random unit quaternion initializer."""

  # Since variable initializers must take `shape` as input, we cannot prevent
  # a clash between util.shape and the argument here. Therefore we have to
  # disable redefined-outer-name for this function.
  # pylint: disable=redefined-outer-name
  def _initializer(shape, dtype=tf.float32, partition_info=None):
    """Generate a random normalized quaternion.

    Note:
      In the following, A1 to An are optional batch dimensions.

    Args:
      shape: A list representing the shape of the output. The last entry of the
        list must be `4`.
      dtype: type of the output (tf.float32 is the only type supported).
      partition_info: how the variable is partitioned (not used).

    Returns:
      A tensor of shape `[A1, ..., An, 4]` representing normalized quaternions.

    Raises:
      ValueError: If `shape` or `dtype` are not supported.
    """
    del partition_info  # unused
    if dtype != tf.float32:
      raise ValueError("'dtype' must be tf.float32.")
    if shape[-1] != 4:
      raise ValueError("Last dimension of 'shape' must be 4.")

    return normalized_random_uniform(shape[:-1])

  return _initializer
  # pylint: enable=redefined-outer-name


def rotate(point, quaternion, name=None):
  """Rotates a point using a quaternion.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    point: A tensor of shape `[A1, ..., An, 3]`, where the last dimension
      represents a 3d point.
    quaternion: A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    name: A name for this op that defaults to "quaternion_rotate".

  Returns:
    A tensor of shape `[A1, ..., An, 3]`, where the last dimension represents a
    3d point.

  Raises:
    ValueError: If the shape of `point` or `quaternion` is not supported.
  """
  with tf.compat.v1.name_scope(name, "quaternion_rotate", [point, quaternion]):
    point = tf.convert_to_tensor(value=point)
    quaternion = tf.convert_to_tensor(value=quaternion)

    shape.check_static(
        tensor=point, tensor_name="point", has_dim_equals=(-1, 3))
    shape.check_static(
        tensor=quaternion, tensor_name="quaternion", has_dim_equals=(-1, 4))
    shape.compare_batch_dimensions(
        tensors=(point, quaternion), last_axes=-2, broadcast_compatible=True)
    quaternion = asserts.assert_normalized(quaternion)

    padding = [[0, 0] for _ in range(point.shape.ndims)]
    padding[-1][-1] = 1
    point = tf.pad(tensor=point, paddings=padding, mode="CONSTANT")
    point = multiply(quaternion, point)
    point = multiply(point, conjugate(quaternion))
    xyz, _ = tf.split(point, (3, 1), axis=-1)
    return xyz


def relative_angle(quaternion1, quaternion2, name=None):
  r"""Computes the unsigned relative rotation angle between 2 unit quaternions.

  Given two normalized quanternions $$\mathbf{q}_1$$ and $$\mathbf{q}_2$$, the
  relative angle is computed as
  $$\theta = 2\arccos(\mathbf{q}_1^T\mathbf{q}_2)$$.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    quaternion1: A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    quaternion2: A tensor of shape `[A1, ..., An, 4]`, where the last dimension
      represents a normalized quaternion.
    name: A name for this op that defaults to "quaternion_relative_angle".

  Returns:
    A tensor of shape `[A1, ..., An, 1]` where the last dimension represents
    rotation angles in the range [0.0, pi].

  Raises:
    ValueError: If the shape of `quaternion1` or `quaternion2` is not supported.
  """
  with (tf.compat.v1.name_scope(name, "quaternion_relative_angle",
                                [quaternion1, quaternion2])):
    quaternion1 = tf.convert_to_tensor(value=quaternion1)
    quaternion2 = tf.convert_to_tensor(value=quaternion2)

    shape.check_static(
        tensor=quaternion1, tensor_name="quaternion1", has_dim_equals=(-1, 4))
    shape.check_static(
        tensor=quaternion2, tensor_name="quaternion2", has_dim_equals=(-1, 4))
    quaternion1 = asserts.assert_normalized(quaternion1)
    quaternion2 = asserts.assert_normalized(quaternion2)

    dot_product = vector.dot(quaternion1, quaternion2, keepdims=False)
    # Ensure dot product is in range [-1. 1].
    eps_dot_prod = 4.0 * asserts.select_eps_for_addition(dot_product.dtype)
    dot_product = safe_ops.safe_shrink(
        dot_product, -1.0, 1.0, False, eps=eps_dot_prod)
    return 2.0 * tf.acos(tf.abs(dot_product))


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trilinear interpolation."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation import grid
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.math.interpolation import trilinear
from tensorflow_graphics.util import test_case


def _sampling_points_from_grid(grid_size, dtype=tf.float64):
  """Returns a tensor of shape `[M, 3]`, with M the number of sampling points."""
  sampling_points = grid.generate((-1.0, -1.0, -1.0), (1.0, 1.0, 1.0),
                                  grid_size)
  sampling_points = tf.cast(sampling_points, dtype)
  return tf.reshape(sampling_points, [-1, 3])


def _transpose_last_two_dims(sampling_points):
  axes = [i for i in range(len(sampling_points.shape))]
  axes[-1], axes[-2] = axes[-2], axes[-1]
  sampling_points = tf.transpose(a=sampling_points, perm=axes)
  return sampling_points


def _sampling_points_in_volume(sampling_points, voxel_size):
  """Transforms the sampling points from [-1, 1] to [0, voxel_size]."""
  voxel_size = tf.convert_to_tensor(value=voxel_size)
  max_size = tf.cast(voxel_size - 1, sampling_points.dtype)
  return 0.5 * ((sampling_points + 1) * max_size)


ANGLE_90 = np.array((np.pi / 2.,))


def _get_random_voxel_grid(voxel_size):
  return np.random.uniform(size=voxel_size)


def _get_random_sampling_points(sampling_points_size, max_grid_dim):
  random_grid = np.random.randint(0, max_grid_dim, size=sampling_points_size)
  return random_grid.astype(np.float64)


def _generate_voxels_horizontal_plane(voxel_size):
  voxels = np.zeros(voxel_size)
  mid_x = int(np.floor(voxel_size[0] / 2))
  voxels[mid_x, :, :, :] = 1
  if voxel_size[0] % 2 == 0:
    voxels[mid_x - 1, :, :, :] = 1
  return voxels


def _generate_voxels_vertical_plane(voxel_size):
  voxels = np.zeros(voxel_size)
  mid_y = int(np.floor(voxel_size[1] / 2))
  voxels[:, mid_y, :, :] = 1
  if voxel_size[1] % 2 == 0:
    voxels[:, mid_y - 1, :, :] = 1
  return voxels


def _generate_voxel_cube(dims, plane_orientation=None):
  if plane_orientation == "horizontal":
    voxels_no_batch = _generate_voxels_horizontal_plane(dims[-4:])
  elif plane_orientation == "vertical":
    voxels_no_batch = _generate_voxels_vertical_plane(dims[-4:])
  else:
    voxels_no_batch = np.zeros(dims[-4:])

  voxels = np.zeros(dims)
  voxels[..., :, :, :, :] = voxels_no_batch
  return voxels


class TrilinearTest(test_case.TestCase):

  @parameterized.parameters(
      ("must have a rank greater than 3", ((5, 5, 5), (125, 3))),
      ("must have a rank greater than 1", ((2, 5, 5, 5, 1), (3,))),
      ("must have exactly 3 dimensions in axis -1", ((2, 5, 5, 5, 1),
                                                     (2, 125, 4))),
      ("Not all batch dimensions are broadcast-compatible.",
       ((2, 2, 5, 5, 5, 1), (2, 3, 125, 3))),
  )
  def test_interpolate_exception_raised(self, error_msg, shapes):
    """Tests whether exceptions are raised for incompatible shapes."""
    self.assert_exception_is_raised(
        trilinear.interpolate, error_msg, shapes=shapes)

  @parameterized.parameters(
      ((5, 5, 5, 3), (125, 3)),
      ((2, 5, 5, 5, 3), (2, 125, 3)),
      ((2, 2, 5, 5, 5, 3), (2, 2, 15, 3)),
  )
  def test_interpolate_exception_not_raised(self, *shapes):
    """Tests whether exceptions are not raised for compatible shapes."""
    self.assert_exception_is_not_raised(trilinear.interpolate, shapes)

  def test_interpolation_values_preset(self):
    voxels = np.zeros((2, 2, 2, 1))
    voxels[(0, 1, 1, 0), (0, 0, 1, 1), (0, 0, 0, 0), 0] = 1
    sampling_points = np.array(((0, 0, 0), (0.5, 0, 0), (1.0, 0, 0),
                                (0., 0, 0.25), (0., 0, 0.5), (0., 0, 0.75),
                                (0., 0, 1.0), (0., 0, 2.0), (-1.0, -0.5, 0),
                                (0, 2, 1.5)))
    correct_values = np.array(
        ((1.0, 1.0, 1.0, 0.75, 0.5, 0.25, 0.0, 0.0, 1.0, 0.0),)).T

    self.assert_output_is_correct(
        trilinear.interpolate, (voxels, sampling_points), (correct_values,),
        tile=False)

  def test_interpolation_preset(self):
    """Tests whether interpolation results are correct."""
    batch_dim_size = np.random.randint(0, 4)
    batch_dims = list(np.random.randint(1, 10, size=batch_dim_size))
    cube_single_dim = np.random.randint(3, 10)
    cube_dims = [cube_single_dim, cube_single_dim, cube_single_dim]
    num_channels = [np.random.randint(1, 10)]
    combined_dims = batch_dims + cube_dims + num_channels
    voxels_in = _generate_voxel_cube(combined_dims, "horizontal")
    euler_angles = np.zeros(batch_dims + [3])
    euler_angles[..., 2] = np.pi / 2.
    voxels_out = _generate_voxel_cube(combined_dims, "vertical")

    transformation_matrix = rotation_matrix_3d.from_euler(euler_angles)
    grid_size = (cube_single_dim, cube_single_dim, cube_single_dim)
    sampling_points = _sampling_points_from_grid(grid_size)
    sampling_points = tf.matmul(transformation_matrix,
                                tf.transpose(a=sampling_points))
    sampling_points = _transpose_last_two_dims(sampling_points)
    sampling_points = _sampling_points_in_volume(sampling_points,
                                                 voxels_in.shape[-4:-1])
    voxels_out = tf.reshape(voxels_out,
                            batch_dims + [cube_single_dim**3] + num_channels)

    self.assert_output_is_correct(
        trilinear.interpolate, (voxels_in, sampling_points), (voxels_out,),
        tile=False)

  @parameterized.parameters(
      (1, 4, 4, 4, 1),
      (2, 4, 4, 4, 3),
      (3, 4, 4, 4, 3),
  )
  def test_interpolate_jacobian_random(self, bsize, height, width, depth,
                                       channels):
    """Tests whether jacobian is correct."""
    grid_3d_np = np.random.uniform(size=(bsize, height, width, depth, channels))
    sampling_points_np = np.zeros((bsize, height * width * depth, 3))
    sampling_points_np[:, :, 0] = np.arange(0, height * width * depth)

    self.assert_jacobian_is_correct_fn(
        lambda grid_3d: trilinear.interpolate(grid_3d, sampling_points_np),
        [grid_3d_np])


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for grid."""

from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.representation import grid
from tensorflow_graphics.util import test_case


class GridTest(test_case.TestCase):

  @parameterized.parameters(
      (((1,), (1,), (1,)), (tf.float32, tf.float32, tf.int32)),
      (((1, 1), (1, 1), (1,)), (tf.float32, tf.float32, tf.int32)),
  )
  def test_generate_exception_not_raised(self, shapes, dtypes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(grid.generate, shapes, dtypes)

  @parameterized.parameters(
      ("starts must have a rank greater than 0", (), (None,), (None,)),
      ("stops must have a rank greater than 0", (None,), (), (None,)),
      ("nums must have a rank of 1", (None,), (None,), ()),
      ("Not all batch dimensions are identical.", (1,), (0,), (1,)),
      ("Not all batch dimensions are identical.", (0,), (1,), (1,)),
      ("must have the same number of dimensions", (1,), (1,), (0,)),
  )
  def test_generate_exception_raised(self, error_msg, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_raised(grid.generate, error_msg, shapes)

  @parameterized.parameters(
      (((-1.,), (1.,), (3,)), (((-1.,), (0.,), (1.,)),)),
      ((((-1.,), (-1.,)), ((1.,), (1.,)), (1,)), ((((-1.,),), ((-1.,),)),)),
  )
  def test_generate_preset(self, test_inputs, test_outputs):
    """Test the uniform grid generation using fix test cases."""
    self.assert_output_is_correct(
        grid.generate, test_inputs, test_outputs, tile=False)

  def test_generate_random(self):
    """Test the uniform grid generation."""
    starts = np.array((0., 0.), dtype=np.float32)
    stops = np.random.randint(1, 10, size=(2))
    nums = stops + 1
    stops = stops.astype(np.float32)

    g = grid.generate(starts, stops, nums)
    shape = nums.tolist() + [2]
    xv, yv = np.meshgrid(range(shape[0]), range(shape[1]), indexing="ij")
    gt = np.stack((xv, yv), axis=-1).astype(np.float32)

    self.assertAllClose(g, gt)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""API export functions used to create the automated documentation."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import inspect


def get_functions_and_classes():
  """Extracts a list of public functions and classes for the API generation.

  Returns:
    A list of function and class names.
  """
  caller = inspect.stack()[1]
  module = inspect.getmodule(caller[0])
  return [
      obj_name for obj_name, obj in inspect.getmembers(module)
      if inspect.isfunction(obj) or
      inspect.isclass(obj) and not obj_name.startswith("_")
  ]


def get_modules():
  """Extracts a list of public modules for the API generation.

  Returns:
    A list of module names.
  """
  caller = inspect.stack()[1]
  module = inspect.getmodule(caller[0])
  return [
      obj_name for obj_name, obj in inspect.getmembers(module)
      if inspect.ismodule(obj) and obj.__name__.rsplit(".", 1)[0] ==
      module.__name__ and not obj_name.startswith("_")
  ]


# The util functions or classes are not exported.
__all__ = []



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Unit test base class.

This class is intended to be used as the unit test base class in TensorFlow
Graphics. It implements new methods on top of the TensorFlow TestCase class
that are used to simplify the code and check for various kinds of failure.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import warnings

from absl import flags
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.util import tfg_flags

FLAGS = flags.FLAGS


def _max_error(arrays1, arrays2):
  """Computes maximum elementwise gap between two lists of ndarrays.

  Computes the maximum elementwise gap between two lists with the same length,
  of arrays with the same shape.

  Args:
    arrays1: a lists of np.ndarrays.
    arrays2: a lists of np.ndarrays of the same shape as arrays1.

  Returns:
    The maximum elementwise absolute difference between the two lists of arrays.
  """
  error = 0
  for array1, array2 in zip(arrays1, arrays2):
    if array1.size or array2.size:  # Handle zero size ndarrays correctly
      error = np.maximum(error, np.fabs(array1 - array2).max())
  return error


class TestCase(parameterized.TestCase, tf.test.TestCase):
  """Test case class implementing extra test functionalities."""

  def setUp(self):  # pylint: disable=invalid-name
    """Sets the seed for tensorflow and numpy."""
    super(TestCase, self).setUp()
    try:
      seed = flags.FLAGS.test_random_seed
    except flags.UnparsedFlagAccessError:
      seed = 301  # Default seed in case test_random_seed is not defined.
    tf.compat.v1.set_random_seed(seed)
    np.random.seed(seed)
    FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value = True

  def _remove_dynamic_shapes(self, shapes):
    for s in shapes:
      if None in s:
        return None
    return shapes

  def _compute_gradient_error(self, x, y, x_init_value, delta=1e-6):
    """Computes the gradient error.

    Args:
      x: a tensor or list of tensors.
      y: a tensor.
      x_init_value: a numpy array of the same shape as "x" representing the
        initial value of x.
      delta: (optional) the amount of perturbation.

    Returns:
      A tuple (max_error, row, column), with max_error the maxium error between
      the two Jacobians, and row/column the position of said maximum error.
    """
    x_shape = x.shape.as_list()
    y_shape = y.shape.as_list()
    with self.cached_session():
      grad = tf.compat.v1.test.compute_gradient(x, x_shape, y, y_shape,
                                                x_init_value, delta)
      if isinstance(grad, tuple):
        grad = [grad]
      error = 0
      row_max_error = 0
      column_max_error = 0
      for j_t, j_n in grad:
        if j_t.size or j_n.size:  # Handle zero size tensors correctly
          diff = np.fabs(j_t - j_n)
          max_error = np.maximum(error, diff.max())
          row_max_error, column_max_error = np.unravel_index(
              diff.argmax(), diff.shape)
      return max_error, row_max_error, column_max_error

  def _create_placeholders_from_shapes(self,
                                       shapes,
                                       dtypes=None,
                                       sparse_tensors=None):
    """Creates a list of placeholders based on a list of shapes.

    Args:
      shapes: A tuple or list of the input shapes.
      dtypes: A list of input types.
      sparse_tensors: A `bool` list denoting if placeholder is a SparseTensor.
        This is ignored in eager mode - in eager execution, only dense
        placeholders will be created.

    Returns:
      A list of placeholders.
    """
    if dtypes is None:
      dtypes = [tf.float32] * len(shapes)
    if sparse_tensors is None:
      sparse_tensors = [False] * len(shapes)
    if tf.executing_eagerly():
      placeholders = [
          tf.compat.v1.placeholder_with_default(
              tf.zeros(shape=shape, dtype=dtype), shape=shape)
          for shape, dtype in zip(shapes, dtypes)
      ]
    else:
      placeholders = [
          tf.compat.v1.sparse.placeholder(dtype, shape=shape)
          if is_sparse else tf.compat.v1.placeholder(shape=shape, dtype=dtype)
          for shape, dtype, is_sparse in zip(shapes, dtypes, sparse_tensors)
      ]
    return placeholders

  def _tile_tensors(self, tiling, tensors):
    """Tiles a set of tensors using the tiling information.

    Args:
      tiling: A list of integers defining how to tile the tensors.
      tensors: A list of tensors to tile.

    Returns:
      A list of tiled tensors.
    """
    tensors = [
        np.tile(tensor, tiling + [1] * len(np.array(tensor).shape))
        for tensor in tensors
    ]
    return tensors

  def assert_exception_is_not_raised(self,
                                     func,
                                     shapes,
                                     dtypes=None,
                                     sparse_tensors=None,
                                     **kwargs):
    """Runs the function to make sure an exception is not raised.

    Args:
      func: A function to exectute.
      shapes: A tuple or list of the input shapes.
      dtypes: A list of input types.
      sparse_tensors: A list of `bool` indicating if the inputs are
        SparseTensors. Defaults to all `False`. This is used for creating
        SparseTensor placeholders in graph mode.
      **kwargs: A dict of keyword arguments to be passed to the function.
    """
    if tf.executing_eagerly() and shapes:
      # If a shape is given in eager mode, the tensor will be initialized with
      # zeros, which can make some range checks fail for certain functions.
      # But if only kwargs are passed and shapes is empty, this function
      # still should run correctly.
      return
    placeholders = self._create_placeholders_from_shapes(
        shapes=shapes, dtypes=dtypes, sparse_tensors=sparse_tensors)
    try:
      func(*placeholders, **kwargs)
    except Exception as e:  # pylint: disable=broad-except
      self.fail("Exception raised: %s" % str(e))

  def assert_exception_is_raised(self,
                                 func,
                                 error_msg,
                                 shapes,
                                 dtypes=None,
                                 sparse_tensors=None,
                                 **kwargs):
    """Runs the function to make sure an exception is raised.

    Args:
      func: A function to exectute.
      error_msg: The error message of the exception.
      shapes: A tuple or list of the input shapes.
      dtypes: A list of input types.
      sparse_tensors: A list of `bool` indicating if the inputs are
        SparseTensors. Defaults to all `False`. This is used for creating
        SparseTensor placeholders in graph mode.
      **kwargs: A dict of keyword arguments to be passed to the function.
    """
    if tf.executing_eagerly():
      # If shapes is an empty list, we can continue with the test. If shapes
      # has None values, we shoud return.
      shapes = self._remove_dynamic_shapes(shapes)
      if shapes is None:
        return
    placeholders = self._create_placeholders_from_shapes(
        shapes=shapes, dtypes=dtypes, sparse_tensors=sparse_tensors)
    with self.assertRaisesRegexp(ValueError, error_msg):
      func(*placeholders, **kwargs)

  def assert_jacobian_is_correct(self, x, x_init, y, atol=1e-6, delta=1e-6):
    """Tests that the gradient error of y=f(x) is small.

    Args:
      x: A tensor.
      x_init: A numpy array containing the values at which to estimate the
        gradients of y.
      y: A tensor.
      atol: Maximum absolute tolerance in gradient error.
      delta: The amount of perturbation.
    """
    warnings.warn((
        "assert_jacobian_is_correct is deprecated and might get "
        "removed in a future version please use assert_jacobian_is_correct_fn"),
                  DeprecationWarning)
    if tf.executing_eagerly():
      self.skipTest(reason="Graph mode only test")
    max_error, _, _ = self._compute_gradient_error(x, y, x_init, delta)
    self.assertLessEqual(max_error, atol)

  def assert_jacobian_is_correct_fn(self, f, x, atol=1e-6, delta=1e-6):
    """Tests that the gradient error of y=f(x) is small.

    Args:
      f: the function.
      x: A list of arguments for the function
      atol: Maximum absolute tolerance in gradient error.
      delta: The amount of perturbation.
    """
    # pylint: disable=no-value-for-parameter
    if tf.executing_eagerly():
      max_error = _max_error(*tf.test.compute_gradient(f, x, delta))
    else:
      with self.cached_session():
        max_error = _max_error(*tf.test.compute_gradient(f, x, delta))
    # pylint: enable=no-value-for-parameter
    self.assertLessEqual(max_error, atol)

  def assert_jacobian_is_finite(self, x, x_init, y):
    """Tests that the Jacobian only contains valid values.

    The analytical gradients and numerical ones are expected to differ at points
    where y is not smooth. This function can be used to check that the
    analytical gradient is not NaN nor Inf.

    Args:
      x: A tensor.
      x_init: A numpy array containing the values at which to estimate the
        gradients of y.
      y: A tensor.
    """
    warnings.warn((
        "assert_jacobian_is_finite is deprecated and might get "
        "removed in a future version please use assert_jacobian_is_finite_fn"),
                  DeprecationWarning)
    if tf.executing_eagerly():
      self.skipTest(reason="Graph mode only test")
    x_shape = x.shape.as_list()
    y_shape = y.shape.as_list()
    with tf.compat.v1.Session():
      gradient = tf.compat.v1.test.compute_gradient(
          x, x_shape, y, y_shape, x_init_value=x_init)
      theoretical_gradient = gradient[0][0]
      self.assertFalse(
          np.isnan(theoretical_gradient).any() or
          np.isinf(theoretical_gradient).any())

  def assert_jacobian_is_finite_fn(self, f, x):
    """Tests that the Jacobian only contains valid values.

    The analytical gradients and numerical ones are expected to differ at points
    where f(x) is not smooth. This function can be used to check that the
    analytical gradient is not 'NaN' nor 'Inf'.

    Args:
      f: the function.
      x: A list of arguments for the function
    """
    if tf.executing_eagerly():
      theoretical_gradient, _ = tf.compat.v2.test.compute_gradient(f, x)
    else:
      with self.cached_session():
        theoretical_gradient, _ = tf.compat.v2.test.compute_gradient(f, x)
    self.assertNotIn(
        True, [
            np.isnan(element).any() or np.isinf(element).any()
            for element in theoretical_gradient
        ],
        msg="nan or inf elements found in theoretical jacobian.")

  def assert_output_is_correct(self,
                               func,
                               test_inputs,
                               test_outputs,
                               rtol=1e-3,
                               atol=1e-6,
                               tile=True):
    """Tests that the function gives the correct result.

    Args:
      func: A function to exectute.
      test_inputs: A tuple or list of test inputs.
      test_outputs: A tuple or list of test outputs against which the result of
        calling `func` on `test_inputs` will be compared to.
      rtol: The relative tolerance used during the comparison.
      atol: The absolute tolerance used during the comparison.
      tile: A `bool` indicating whether or not to automatically tile the test
        inputs and outputs.
    """
    if tile:
      # Creates a rank 4 list of values between 1 and 10.
      tensor_tile = np.random.randint(1, 10, size=np.random.randint(4)).tolist()
      test_inputs = self._tile_tensors(tensor_tile, test_inputs)
      test_outputs = self._tile_tensors(tensor_tile, test_outputs)
    test_outputs = [
        tf.convert_to_tensor(value=output) for output in test_outputs
    ]
    test_outputs = test_outputs[0] if len(test_outputs) == 1 else test_outputs
    self.assertAllClose(test_outputs, func(*test_inputs), rtol=rtol, atol=atol)

  def assert_tf_lite_convertible(self,
                                 func,
                                 shapes,
                                 dtypes=None,
                                 test_inputs=None):
    """Runs the tf-lite converter to make sure the function can be exported.

    Args:
      func: A function to execute with tf-lite.
      shapes: A tuple or list of input shapes.
      dtypes: A list of input types.
      test_inputs: A tuple or list of inputs. If not provided the test inputs
        will be randomly generated.
    """
    if tf.executing_eagerly():
      # Currently TFLite conversion is not supported in eager mode.
      self.skipTest(reason="Graph mode only test")
    # Generate graph with the function given as input.
    in_tensors = self._create_placeholders_from_shapes(shapes, dtypes)
    out_tensors = func(*in_tensors)
    if not isinstance(out_tensors, (list, tuple)):
      out_tensors = [out_tensors]
    with tf.compat.v1.Session() as sess:
      try:
        sess.run(tf.compat.v1.global_variables_initializer())
        # Convert to a TFLite model.
        converter = tf.compat.v1.lite.TFLiteConverter.from_session(
            sess, in_tensors, out_tensors)
        tflite_model = converter.convert()
        # Load TFLite model and allocate tensors.
        interpreter = tf.lite.Interpreter(model_content=tflite_model)
        interpreter.allocate_tensors()
        # If no test inputs provided then randomly generate inputs.
        if test_inputs is None:
          test_inputs = [
              np.array(np.random.sample(shape), dtype=np.float32)
              for shape in shapes
          ]
        else:
          test_inputs = [
              np.array(test, dtype=np.float32) for test in test_inputs
          ]
        # Evaluate function using TensorFlow.
        feed_dict = dict(zip(in_tensors, test_inputs))
        test_outputs = sess.run(out_tensors, feed_dict)
        # Set tensors for the TFLite model.
        input_details = interpreter.get_input_details()
        for i, test_input in enumerate(test_inputs):
          index = input_details[i]["index"]
          interpreter.set_tensor(index, test_input)
        # Run TFLite model.
        interpreter.invoke()
        # Get tensors from the TFLite model and compare with TensorFlow.
        output_details = interpreter.get_output_details()
        for o, test_output in enumerate(test_outputs):
          index = output_details[o]["index"]
          self.assertAllClose(test_output, interpreter.get_tensor(index))
      except Exception as e:  # pylint: disable=broad-except
        self.fail("Exception raised: %s" % str(e))


def main(argv=None):
  """Main function."""
  tf.test.main(argv)


# The util functions or classes are not exported.
__all__ = []



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""This module implements utility functions for the linear RGB color space."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import tensorflow as tf
from tensorflow_graphics.image.color_space import constants
from tensorflow_graphics.util import asserts
from tensorflow_graphics.util import export_api
from tensorflow_graphics.util import shape

# Conversion constants following the naming convention from the 'theory of the
# transformation' section at https://en.wikipedia.org/wiki/SRGB.
_A = constants.srgb_gamma["A"]
_PHI = constants.srgb_gamma["PHI"]
_K0 = constants.srgb_gamma["K0"]
_GAMMA = constants.srgb_gamma["GAMMA"]


def from_srgb(srgb, name=None):
  """Converts sRGB colors to linear colors.

  Note:
      In the following, A1 to An are optional batch dimensions.

  Args:
    srgb: A tensor of shape `[A_1, ..., A_n, 3]`, where the last dimension
      represents sRGB values.
    name: A name for this op that defaults to "srgb_to_linear".

  Raises:
    ValueError: If `srgb` has rank < 1 or has its last dimension not equal to 3.

  Returns:
    A tensor of shape `[A_1, ..., A_n, 3]`, where the last dimension represents
    RGB values in linear color space.
  """
  with tf.compat.v1.name_scope(name, "linear_rgb_from_srgb", [srgb]):
    srgb = tf.convert_to_tensor(value=srgb)

    shape.check_static(
        tensor=srgb,
        tensor_name="srgb",
        has_rank_greater_than=0,
        has_dim_equals=(-1, 3))

    srgb = asserts.assert_all_in_range(srgb, 0., 1.)
    return tf.compat.v1.where(srgb <= _K0, srgb / _PHI,
                              ((srgb + _A) / (1 + _A))**_GAMMA)


# API contains all public functions and classes.
__all__ = export_api.get_functions_and_classes()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for srgb."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np

from tensorflow_graphics.image.color_space import linear_rgb
from tensorflow_graphics.image.color_space import srgb
from tensorflow_graphics.util import test_case


class LinearRGBTest(test_case.TestCase):

  def test_cycle_srgb_linear_rgb_srgb_for_random_input(self):
    """Tests loop from sRGB to linear RGB and back for random inputs."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    srgb_input = np.random.uniform(size=tensor_shape + [3])
    linear_output = linear_rgb.from_srgb(srgb_input)
    srgb_recovered = srgb.from_linear_rgb(linear_output)

    self.assertAllClose(srgb_input, srgb_recovered)

  @parameterized.parameters(
      (((0., 0.5, 1.), (0.0404, 0.04045, 0.0405)),
       ((0., 0.214041, 1.), (0.003127, 0.003131, 0.003135))),)
  def test_from_srgb_preset(self, test_inputs, test_outputs):
    """Tests conversion from sRGB to linear RGB space for preset inputs."""
    self.assert_output_is_correct(linear_rgb.from_srgb, (test_inputs,),
                                  (test_outputs,))

  def test_from_srgb_jacobian_random(self):
    """Tests the Jacobian of the from_srgb function for random inputs."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    srgb_random_init = np.random.uniform(size=tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(linear_rgb.from_srgb, [srgb_random_init])

  @parameterized.parameters(
      (np.array((0., 0.01, 0.02)),), (np.array((0.05, 0.06, 1.)),),
      (np.array((0.01, 0.04, 0.06)),))
  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_srgb_jacobian_preset(self, inputs_init):
    """Tests the Jacobian of the from_srgb function for preset inputs."""
    self.assert_jacobian_is_correct_fn(linear_rgb.from_srgb, [inputs_init])

  @parameterized.parameters(
      ((3,),),
      ((None, None, None, 3),),
  )
  def test_from_srgb_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(linear_rgb.from_srgb, shape)

  @parameterized.parameters(
      ("must have a rank greater than 0", ()),
      ("must have exactly 3 dimensions in axis -1", (2, 3, 4)),
  )
  def test_from_srgb_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are properly raised."""
    self.assert_exception_is_raised(linear_rgb.from_srgb, error_msg, shape)


if __name__ == "__main__":
  test_case.main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Assert functions to be used by various modules.

This module contains asserts that are intended to be used in TensorFlow
Graphics. These asserts will be activated only if the debug flag
TFG_ADD_ASSERTS_TO_GRAPH is set to True.
"""

from absl import flags
import numpy as np
import tensorflow as tf

from tensorflow_graphics.util import tfg_flags

FLAGS = flags.FLAGS


def assert_no_infs_or_nans(tensor, name=None):
  """Checks a tensor for NaN and Inf values.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    tensor: A tensor of shape `[A1, ..., An]` containing the values we want to
      check.
    name: A name for this op. Defaults to 'assert_no_infs_or_nans'.

  Raises:
    tf.errors.InvalidArgumentError: If any entry of the input is NaN or Inf.

  Returns:
    The input vector, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return tensor

  with tf.compat.v1.name_scope(name, 'assert_no_infs_or_nans', [tensor]):
    tensor = tf.convert_to_tensor(value=tensor)

    assert_ops = (tf.debugging.check_numerics(
        tensor, message='Inf or NaN detected.'),)
    with tf.control_dependencies(assert_ops):
      return tf.identity(tensor)


def assert_all_above(vector, minval, open_bound=False, name=None):
  """Checks whether all values of vector are above minval.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector: A tensor of shape `[A1, ..., An]` containing the values we want to
      check.
    minval: A scalar or a tensor of shape `[A1, ..., An]` representing the
      desired lower bound for the values in `vector`.
    open_bound: A `bool` indicating whether the range is open or closed.
    name: A name for this op. Defaults to 'assert_all_above'.

  Raises:
    tf.errors.InvalidArgumentError: If any entry of the input is below `minval`.

  Returns:
    The input vector, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return vector

  with tf.compat.v1.name_scope(name, 'assert_all_above', [vector, minval]):
    vector = tf.convert_to_tensor(value=vector)
    minval = tf.convert_to_tensor(value=minval, dtype=vector.dtype)

    if open_bound:
      assert_ops = (tf.compat.v1.assert_greater(vector, minval),)
    else:
      assert_ops = (tf.compat.v1.assert_greater_equal(vector, minval),)
    with tf.control_dependencies(assert_ops):
      return tf.identity(vector)


def assert_all_below(vector, maxval, open_bound=False, name=None):
  """Checks whether all values of vector are below maxval.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector: A tensor of shape `[A1, ..., An]` containing the values we want to
      check.
    maxval: A scalar or a tensor of shape `[A1, ..., An]` representing the
      desired upper bound for the values in `vector`.
    open_bound: A boolean indicating whether the range is open or closed.
    name: A name for this op. Defaults to 'assert_all_below'.

  Raises:
    tf.errors.InvalidArgumentError: If any entry of the input exceeds `maxval`.

  Returns:
    The input vector, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return vector

  with tf.compat.v1.name_scope(name, 'assert_all_below', [vector, maxval]):
    vector = tf.convert_to_tensor(value=vector)
    maxval = tf.convert_to_tensor(value=maxval, dtype=vector.dtype)

    if open_bound:
      assert_ops = (tf.compat.v1.assert_less(vector, maxval),)
    else:
      assert_ops = (tf.compat.v1.assert_less_equal(vector, maxval),)
    with tf.control_dependencies(assert_ops):
      return tf.identity(vector)


def assert_all_in_range(vector, minval, maxval, open_bounds=False, name=None):
  """Checks whether all values of vector are between minval and maxval.

  This function checks if all the values in the given vector are in an interval
  `[minval, maxval]` if `open_bounds` is `False`, or in `]minval, maxval[` if it
  is set to `True`.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector: A tensor of shape `[A1, ..., An]` containing the values we want to
      check.
    minval: A `float` or a tensor of shape `[A1, ..., An]` representing the
      desired lower bound for the values in `vector`.
    maxval: A `float` or a tensor of shape `[A1, ..., An]` representing the
      desired upper bound for the values in `vector`.
    open_bounds: A `bool` indicating whether the range is open or closed.
    name: A name for this op. Defaults to 'assert_all_in_range'.

  Raises:
    tf.errors.InvalidArgumentError: If `vector` is not in the expected range.

  Returns:
    The input vector, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return vector

  with tf.compat.v1.name_scope(name, 'assert_all_in_range',
                               [vector, minval, maxval]):
    vector = tf.convert_to_tensor(value=vector)
    minval = tf.convert_to_tensor(value=minval, dtype=vector.dtype)
    maxval = tf.convert_to_tensor(value=maxval, dtype=vector.dtype)

    if open_bounds:
      assert_ops = (tf.compat.v1.assert_less(vector, maxval),
                    tf.compat.v1.assert_greater(vector, minval))
    else:
      assert_ops = (tf.compat.v1.assert_less_equal(vector, maxval),
                    tf.compat.v1.assert_greater_equal(vector, minval))
    with tf.control_dependencies(assert_ops):
      return tf.identity(vector)


def assert_nonzero_norm(vector, eps=None, name=None):
  """Checks whether vector/quaternion has non-zero norm in its last dimension.

  This function checks whether all the norms of the vectors are greater than
  eps, such that normalizing them will not generate NaN values. Normalization is
  assumed to be done in the last dimension of vector. If eps is left as `None`,
  the function will determine the most suitable value depending on the `dtype`
  of the `vector`.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector: A tensor of shape `[A1, ..., An, V]`, where the last dimension
      contains a V dimensional vector.
    eps: A `float` describing the tolerance used to determine if the norm is
      equal to zero.
    name: A name for this op. Defaults to 'assert_nonzero_norm'.

  Raises:
    InvalidArgumentError: If `vector` has zero norm.

  Returns:
    The input vector, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return vector

  with tf.compat.v1.name_scope(name, 'assert_nonzero_norm', [vector, eps]):
    vector = tf.convert_to_tensor(value=vector)
    if eps is None:
      eps = select_eps_for_division(vector.dtype)
    eps = tf.convert_to_tensor(value=eps, dtype=vector.dtype)

    norm = tf.norm(tensor=vector, axis=-1)
    with tf.control_dependencies([tf.compat.v1.assert_greater(norm, eps)]):
      return tf.identity(vector)


def assert_normalized(vector, order='euclidean', axis=-1, eps=None, name=None):
  """Checks whether vector/quaternion is normalized in its last dimension.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    vector: A tensor of shape `[A1, ..., M, ..., An]`, where the axis of M
      contains the vectors.
    order: Order of the norm passed to tf.norm.
    axis: The axis containing the vectors.
    eps: A `float` describing the tolerance used to determine if the norm is
      equal to `1.0`.
    name: A name for this op. Defaults to 'assert_normalized'.

  Raises:
    InvalidArgumentError: If the norm of `vector` is not `1.0`.

  Returns:
    The input vector, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return vector

  with tf.compat.v1.name_scope(name, 'assert_normalized', [vector]):
    vector = tf.convert_to_tensor(value=vector)
    if eps is None:
      eps = select_eps_for_division(vector.dtype)
    eps = tf.convert_to_tensor(value=eps, dtype=vector.dtype)

    norm = tf.norm(tensor=vector, ord=order, axis=axis)
    one = tf.constant(1.0, dtype=norm.dtype)
    with tf.control_dependencies(
        [tf.compat.v1.assert_near(norm, one, atol=eps)]):
      return tf.identity(vector)


def assert_at_least_k_non_zero_entries(tensor, k=1, name=None):
  """Checks if `tensor` has at least k non-zero entries in the last dimension.

  Given a tensor with `M` dimensions in its last axis, this function checks
  whether at least `k` out of `M` dimensions are non-zero.

  Note:
    In the following, A1 to An are optional batch dimensions.

  Args:
    tensor: A tensor of shape `[A1, ..., An, M]`.
    k: An integer, corresponding to the minimum number of non-zero entries.
    name: A name for this op. Defaults to 'assert_at_least_k_non_zero_entries'.

  Raises:
    InvalidArgumentError: If `tensor` has less than `k` non-zero entries in its
      last axis.

  Returns:
    The input tensor, with dependence on the assertion operator in the graph.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return tensor

  with tf.compat.v1.name_scope(name, 'assert_at_least_k_non_zero_entries',
                               [tensor, k]):
    tensor = tf.convert_to_tensor(value=tensor)

    indicator = tf.cast(tf.math.greater(tensor, 0.0), dtype=tensor.dtype)
    indicator_sum = tf.reduce_sum(input_tensor=indicator, axis=-1)
    assert_op = tf.compat.v1.assert_greater_equal(
        indicator_sum, tf.cast(k, dtype=tensor.dtype))
    with tf.control_dependencies([assert_op]):
      return tf.identity(tensor)


def assert_binary(tensor, name=None):
  """Asserts that all the values in the tensor are zeros or ones.

  Args:
    tensor: A tensor of shape `[A1, ..., An]` containing the values we want to
      check.
    name: A name for this op. Defaults to "assert_binary".

  Returns:
    The input tensor, with dependence on the assertion operator in the graph.

  Raises:
    tf.errors.InvalidArgumentError: If any of the values in the tensor is not
    zero or one.
  """
  if not FLAGS[tfg_flags.TFG_ADD_ASSERTS_TO_GRAPH].value:
    return tensor

  with tf.compat.v1.name_scope(name, 'assert_binary', [tensor]):
    tensor = tf.convert_to_tensor(value=tensor)
    condition = tf.reduce_all(
        input_tensor=tf.logical_or(tf.equal(tensor, 0), tf.equal(tensor, 1)))

    with tf.control_dependencies([tf.Assert(condition, data=[tensor])]):
      return tf.identity(tensor)


def select_eps_for_addition(dtype):
  """Returns 2 * machine epsilon based on `dtype`.

  This function picks an epsilon slightly greater than the machine epsilon,
  which is the upper bound on relative error. This value ensures that
  `1.0 + eps != 1.0`.

  Args:
    dtype: The `tf.DType` of the tensor to which eps will be added.

  Raises:
    ValueError: If `dtype` is not a floating type.

  Returns:
    A `float` to be used to make operations safe.
  """
  return 2.0 * np.finfo(dtype.as_numpy_dtype).eps


def select_eps_for_division(dtype):
  """Selects default values for epsilon to make divisions safe based on dtype.

  This function returns an epsilon slightly greater than the smallest positive
  floating number that is representable for the given dtype. This is mainly used
  to prevent division by zero, which produces Inf values. However, if the
  nominator is orders of magnitude greater than `1.0`, eps should also be
  increased accordingly. Only floating types are supported.

  Args:
    dtype: The `tf.DType` of the tensor to which eps will be added.

  Raises:
    ValueError: If `dtype` is not a floating type.

  Returns:
    A `float` to be used to make operations safe.
  """
  return 10.0 * np.finfo(dtype.as_numpy_dtype).tiny


# The util functions or classes are not exported.
__all__ = []



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Tests the ModelNet40 dataset with fake data."""

import os
import tensorflow_datasets as tfds
from tensorflow_graphics.datasets import modelnet40


class ModelNet40Test(tfds.testing.DatasetBuilderTestCase):
  """Tests the ModelNet40 dataset with fake data."""
  DATASET_CLASS = modelnet40.ModelNet40
  SPLITS = {
      "train": 24,  # Number of fake train example
      "test": 16,  # Number of fake test example
  }

  # If you are calling `download/download_and_extract` with a dict, like:
  #   dl_manager.download({'some_key': 'http://a.org/out.txt', ...})
  # then the tests needs to provide the fake output paths relative to the
  # fake data directory
  DL_EXTRACT_RESULT = ""
  EXAMPLE_DIR = os.path.join(os.path.dirname(__file__), "fakes")
  # SKIP_CHECKSUMS = True


if __name__ == "__main__":
  tfds.testing.test_main()



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# Lint as: python3
"""Voxel grid feature."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import numpy as np
from scipy import io as sio
import tensorflow as tf
from tensorflow_datasets import features


class VoxelGrid(features.Tensor):
  """`FeatureConnector` for voxel grids.

  During `_generate_examples`, the feature connector accepts as input any of:

    * `dict`: dictionary containing the path to a {.mat} file and the key under
    which the voxel grid is accessible inside the file.
    Structure of the dictionary:
      {
      'path': 'path/to/file.mat',
      'key': 'foo'
      }

    * `np.ndarray`: float32 numpy array of shape [X,Y,Z] representing the
    voxel grid.

  Output:
    A float32 Tensor with shape [X,Y,Z] containing the voxel occupancies.
  """

  def __init__(self, shape):
    super(VoxelGrid, self).__init__(shape=shape, dtype=tf.float32)

  def encode_example(self, example_data):
    # Path to .mat file
    if isinstance(example_data, dict):

      if not all(key in example_data for key in ['path', 'key']):
        raise ValueError(
            f'Missing keys in provided dictionary! Expecting \'path\''
            f' and \'key\', but {example_data.keys()} were given.')

      if not os.path.exists(example_data['path']):
        raise FileNotFoundError(
            f"File `{example_data['path']}` does not exist.")

      with tf.io.gfile.GFile(example_data['path'], 'rb') as mat_file:
        voxel_mat = sio.loadmat(mat_file)

      if example_data['key'] not in voxel_mat:
        raise ValueError(f"Key `{example_data['key']}` not found in .mat file. "
                         f"Available keys in file: {voxel_mat.keys()}")

      voxel_grid = voxel_mat[example_data['key']].astype(np.float32)

    else:
      if example_data.ndim != 3:
        raise ValueError('Only 3D Voxel Grids are supported.')

      voxel_grid = example_data

    return super(VoxelGrid, self).encode_example(voxel_grid)



# Copyright 2020 The TensorFlow Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for quaternion."""

from absl.testing import flagsaver
from absl.testing import parameterized
import numpy as np
import tensorflow as tf

from tensorflow_graphics.geometry.transformation import axis_angle
from tensorflow_graphics.geometry.transformation import quaternion
from tensorflow_graphics.geometry.transformation import rotation_matrix_3d
from tensorflow_graphics.geometry.transformation.tests import test_data as td
from tensorflow_graphics.geometry.transformation.tests import test_helpers
from tensorflow_graphics.util import test_case


class QuaternionTest(test_case.TestCase):

  @parameterized.parameters(
      ((3,), (3,)),
      ((None, 3), (None, 3)),
  )
  def test_between_two_vectors_3d_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(quaternion.between_two_vectors_3d,
                                        shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (2,), (3,)),
      ("must have exactly 3 dimensions", (3,), (2,)),
  )
  def test_between_two_vectors_3d_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.between_two_vectors_3d,
                                    error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_between_two_vectors_3d_jacobian_random(self):
    """Tests the Jacobian of between_two_vectors_3d."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    x_1_init = np.random.random(tensor_shape + [3])
    x_2_init = np.random.random(tensor_shape + [3])

    self.assert_jacobian_is_correct_fn(
        quaternion.between_two_vectors_3d, [x_1_init, x_2_init], atol=1e-4)

  def test_between_two_vectors_3d_random(self):
    """Checks the extracted rotation between two 3d vectors."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()
    source = np.random.random(tensor_shape + [3]).astype(np.float32)
    target = np.random.random(tensor_shape + [3]).astype(np.float32)

    rotation = quaternion.between_two_vectors_3d(source, target)
    rec_target = quaternion.rotate(source, rotation)

    self.assertAllClose(
        tf.nn.l2_normalize(target, axis=-1),
        tf.nn.l2_normalize(rec_target, axis=-1))
    # Checks that resulting quaternions are normalized.
    self.assertAllEqual(
        quaternion.is_normalized(rotation), np.full(tensor_shape + [1], True))

  def test_between_two_vectors_3d_that_are_the_same(self):
    """Checks the extracted rotation between two identical 3d vectors."""
    source = np.random.random((1, 3))
    rotation = quaternion.between_two_vectors_3d(source, source)
    self.assertAllEqual([[0, 0, 0, 1]], rotation)

  def test_between_two_vectors_3d_that_are_collinear(self):
    """Checks the extracted rotation between two collinear 3d vectors."""
    axis = [(1.0, 0.0, 0.0), (0.0, 1.0, 0.0)]
    antiparallel_axis = [(0.0, 1.0, 0.0), (0.0, 0.0, 1.0)]
    source = np.multiply(axis, 10.)
    target = np.multiply(axis, -10.)
    rotation = quaternion.between_two_vectors_3d(source, target)
    rotation_pi = quaternion.from_axis_angle(antiparallel_axis,
                                             [[np.pi], [np.pi]])
    self.assertAllClose(rotation_pi, rotation)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
  )
  def test_conjugate_exception_not_raised(self, *shape):
    """Tests that the shape exceptions of conjugate are not raised."""
    self.assert_exception_is_not_raised(quaternion.conjugate, shape)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (3,)),)
  def test_conjugate_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.conjugate, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_conjugate_jacobian_preset(self):
    """Test the Jacobian of the conjugate function."""
    x_init = test_helpers.generate_preset_test_quaternions()

    self.assert_jacobian_is_correct_fn(quaternion.conjugate, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_conjugate_jacobian_random(self):
    """Test the Jacobian of the conjugate function."""
    x_init = test_helpers.generate_random_test_quaternions()

    self.assert_jacobian_is_correct_fn(quaternion.conjugate, [x_init])

  @parameterized.parameters(
      ((3,), (1,)),
      ((None, 3), (None, 1)),
  )
  def test_from_axis_angle_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(quaternion.from_axis_angle, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (2,), (1,)),
      ("must have exactly 1 dimensions", (3,), (2,)),
  )
  def test_from_axis_angle_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.from_axis_angle, error_msg,
                                    shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_axis_angle_jacobian_preset(self):
    """Test the Jacobian of the from_axis_angle function."""
    x_axis_init, x_angle_init = test_helpers.generate_preset_test_axis_angle()

    self.assert_jacobian_is_correct_fn(quaternion.from_axis_angle,
                                       [x_axis_init, x_angle_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_axis_angle_jacobian_random(self):
    """Test the Jacobian of the from_axis_angle function."""
    x_axis_init, x_angle_init = test_helpers.generate_random_test_axis_angle()

    self.assert_jacobian_is_correct_fn(quaternion.from_axis_angle,
                                       [x_axis_init, x_angle_init])

  def test_from_axis_angle_normalized_random(self):
    """Test that from_axis_angle produces normalized quaternions."""
    random_axis, random_angle = test_helpers.generate_random_test_axis_angle()

    random_quaternion = quaternion.from_axis_angle(random_axis, random_angle)

    self.assertAllEqual(
        quaternion.is_normalized(random_quaternion),
        np.ones(shape=random_angle.shape, dtype=bool))

  def test_from_axis_angle_random(self):
    """Tests converting an axis-angle to a quaternion."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    axis, angle = axis_angle.from_euler(random_euler_angles)
    grountruth = rotation_matrix_3d.from_quaternion(
        quaternion.from_euler(random_euler_angles))
    prediction = rotation_matrix_3d.from_quaternion(
        quaternion.from_axis_angle(axis, angle))

    self.assertAllClose(grountruth, prediction, rtol=1e-3)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_from_euler_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(quaternion.from_euler, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (4,)),)
  def test_from_euler_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.from_euler, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_preset(self):
    """Test the Jacobian of the from_euler function."""
    x_init = test_helpers.generate_preset_test_euler_angles()

    self.assert_jacobian_is_correct_fn(quaternion.from_euler, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_euler_jacobian_random(self):
    """Test the Jacobian of the from_euler function."""
    x_init = test_helpers.generate_random_test_euler_angles()

    self.assert_jacobian_is_correct_fn(quaternion.from_euler, [x_init])

  def test_from_euler_normalized_random(self):
    """Tests that quaternions.from_euler returns normalized quaterions."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()
    tensor_shape = random_euler_angles.shape[:-1]

    random_quaternion = quaternion.from_euler(random_euler_angles)

    self.assertAllEqual(
        quaternion.is_normalized(random_quaternion),
        np.ones(shape=tensor_shape + (1,), dtype=bool))

  def test_from_euler_random(self):
    """Tests that quaternions can be constructed from Euler angles."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()
    tensor_shape = random_euler_angles.shape[:-1]

    random_matrix = rotation_matrix_3d.from_euler(random_euler_angles)
    random_quaternion = quaternion.from_euler(random_euler_angles)
    random_point = np.random.normal(size=tensor_shape + (3,))
    rotated_with_matrix = rotation_matrix_3d.rotate(random_point, random_matrix)
    rotated_with_quaternion = quaternion.rotate(random_point, random_quaternion)

    self.assertAllClose(rotated_with_matrix, rotated_with_quaternion)

  @parameterized.parameters(
      ((3,),),
      ((None, 3),),
  )
  def test_from_euler_with_small_angles_approximation_exception_not_raised(
      self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(
        quaternion.from_euler_with_small_angles_approximation, shapes)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (4,)),)
  def test_from_euler_with_small_angles_approximation_exception_raised(
      self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(
        quaternion.from_euler_with_small_angles_approximation, error_msg, shape)

  def test_from_euler_with_small_angles_approximation_random(self):
    # Only generate small angles. For a test tolerance of 1e-3, 0.33 was found
    # empirically to be the range where the small angle approximation works.
    random_euler_angles = test_helpers.generate_random_test_euler_angles(
        min_angle=-0.33, max_angle=0.33)

    exact_quaternion = quaternion.from_euler(random_euler_angles)
    approximate_quaternion = (
        quaternion.from_euler_with_small_angles_approximation(
            random_euler_angles))

    self.assertAllClose(exact_quaternion, approximate_quaternion, atol=1e-3)

  @parameterized.parameters(
      ("must have a rank greater than 1", (3,)),
      ("must have exactly 3 dimensions", (4, 3)),
      ("must have exactly 3 dimensions", (3, 4)),
  )
  def test_from_rotation_matrix_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.from_rotation_matrix, error_msg,
                                    shape)

  @parameterized.parameters(
      ((3, 3),),
      ((None, 3, 3),),
  )
  def test_from_rotation_matrix_not_raised(self, *shape):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(quaternion.from_rotation_matrix, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_rotation_matrix_jacobian_preset(self):
    """Test the Jacobian of the from_rotation_matrix function."""
    x_init = test_helpers.generate_preset_test_rotation_matrices_3d()
    x = tf.convert_to_tensor(value=x_init)

    y = quaternion.from_rotation_matrix(x)

    self.assert_jacobian_is_finite(x, x_init, y)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_from_rotation_matrix_jacobian_random(self):
    """Test the Jacobian of the from_rotation_matrix function."""
    x_init = test_helpers.generate_random_test_rotation_matrix_3d()

    self.assert_jacobian_is_finite_fn(quaternion.from_rotation_matrix, [x_init])

  def test_from_rotation_matrix_normalized_random(self):
    """Tests that from_rotation_matrix produces normalized quaternions."""
    random_matrix = test_helpers.generate_random_test_rotation_matrix_3d()

    random_quaternion = quaternion.from_rotation_matrix(random_matrix)

    self.assertAllEqual(
        quaternion.is_normalized(random_quaternion),
        np.ones(shape=random_matrix.shape[:-2] + (1,), dtype=bool))

  @parameterized.parameters(
      ((td.MAT_3D_ID,), (td.QUAT_ID,)),
      ((td.MAT_3D_X_45,), (td.QUAT_X_45,)),
      ((td.MAT_3D_Y_45,), (td.QUAT_Y_45,)),
      ((td.MAT_3D_Z_45,), (td.QUAT_Z_45,)),
      ((td.MAT_3D_X_90,), (td.QUAT_X_90,)),
      ((td.MAT_3D_Y_90,), (td.QUAT_Y_90,)),
      ((td.MAT_3D_Z_90,), (td.QUAT_Z_90,)),
      ((td.MAT_3D_X_180,), (td.QUAT_X_180,)),
      ((td.MAT_3D_Y_180,), (td.QUAT_Y_180,)),
      ((td.MAT_3D_Z_180,), (td.QUAT_Z_180,)),
  )
  def test_from_rotation_matrix_preset(self, test_inputs, test_outputs):
    self.assert_output_is_correct(quaternion.from_rotation_matrix, test_inputs,
                                  test_outputs)

  def test_from_rotation_matrix_random(self):
    """Tests that from_rotation_matrix produces the expected quaternions."""
    random_euler_angles = test_helpers.generate_random_test_euler_angles()

    random_rotation_matrix_3d = rotation_matrix_3d.from_euler(
        random_euler_angles)
    groundtruth = rotation_matrix_3d.from_quaternion(
        quaternion.from_euler(random_euler_angles))
    prediction = rotation_matrix_3d.from_quaternion(
        quaternion.from_rotation_matrix(random_rotation_matrix_3d))

    self.assertAllClose(groundtruth, prediction)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
  )
  def test_inverse_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_not_raised(quaternion.inverse, shape)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (3,)),)
  def test_inverse_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.inverse, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_preset(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_preset_test_quaternions()

    self.assert_jacobian_is_correct_fn(quaternion.inverse, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_inverse_jacobian_random(self):
    """Test the Jacobian of the inverse function."""
    x_init = test_helpers.generate_random_test_quaternions()

    self.assert_jacobian_is_correct_fn(quaternion.inverse, [x_init])

  def test_inverse_normalized_random(self):
    """Tests that the inverse function returns normalized quaternions."""
    random_quaternion = test_helpers.generate_random_test_quaternions()

    inverse_quaternion = quaternion.inverse(random_quaternion)

    self.assertAllEqual(
        quaternion.is_normalized(inverse_quaternion),
        np.ones(shape=random_quaternion.shape[:-1] + (1,), dtype=bool))

  def test_inverse_random(self):
    """Tests that multiplying with the inverse gives identity."""
    random_quaternion = test_helpers.generate_random_test_quaternions()

    inverse_quaternion = quaternion.inverse(random_quaternion)
    final_quaternion = quaternion.multiply(random_quaternion,
                                           inverse_quaternion)
    tensor_shape = random_quaternion.shape[:-1]
    identity_quaternion = np.array((0.0, 0.0, 0.0, 1.0), dtype=np.float32)
    identity_quaternion = np.tile(identity_quaternion, tensor_shape + (1,))

    self.assertAllClose(final_quaternion, identity_quaternion, rtol=1e-3)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
  )
  def test_is_normalized_exception_not_raised(self, *shape):
    """Tests that the shape exceptions of from_quaternion are not raised."""
    self.assert_exception_is_not_raised(quaternion.is_normalized, shape)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (1, 5)),)
  def test_is_normalized_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions of from_quaternion are raised."""
    self.assert_exception_is_raised(quaternion.is_normalized, error_msg, shape)

  def test_is_normalized_random(self):
    """Tests that is_normalized works as intended."""
    random_quaternion = test_helpers.generate_random_test_quaternions()
    tensor_shape = random_quaternion.shape[:-1]

    unnormalized_random_quaternion = random_quaternion * 1.01
    quat = np.concatenate((random_quaternion, unnormalized_random_quaternion),
                          axis=0)
    mask = np.concatenate(
        (np.ones(shape=tensor_shape + (1,),
                 dtype=bool), np.zeros(shape=tensor_shape + (1,), dtype=bool)),
        axis=0)
    is_normalized = quaternion.is_normalized(quat)

    self.assertAllEqual(mask, is_normalized)

  @parameterized.parameters(
      ((4,),),
      ((None, 4),),
  )
  def test_normalize_exception_not_raised(self, *shape):
    """Tests that the shape exceptions of from_quaternion are not raised."""
    self.assert_exception_is_not_raised(quaternion.normalize, shape)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (1, 5)),)
  def test_normalize_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions of from_quaternion are raised."""
    self.assert_exception_is_raised(quaternion.normalize, error_msg, shape)

  def test_normalize_random(self):
    """Tests that normalize works as intended."""
    random_quaternion = test_helpers.generate_random_test_quaternions()
    tensor_shape = random_quaternion.shape[:-1]

    unnormalized_random_quaternion = random_quaternion * 1.01
    quat = np.concatenate((random_quaternion, unnormalized_random_quaternion),
                          axis=0)
    mask = np.concatenate(
        (np.ones(shape=tensor_shape + (1,),
                 dtype=bool), np.zeros(shape=tensor_shape + (1,), dtype=bool)),
        axis=0)
    is_normalized_before = quaternion.is_normalized(quat)
    normalized = quaternion.normalize(quat)
    is_normalized_after = quaternion.is_normalized(normalized)

    self.assertAllEqual(mask, is_normalized_before)
    self.assertAllEqual(is_normalized_after,
                        np.ones(shape=is_normalized_after.shape, dtype=bool))

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_normalize_jacobian_preset(self):
    """Test the Jacobian of the normalize function."""
    x_init = test_helpers.generate_preset_test_quaternions()

    self.assert_jacobian_is_correct_fn(quaternion.normalize, [x_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_normalize_jacobian_random(self):
    """Test the Jacobian of the normalize function."""
    tensor_dimensions = np.random.randint(low=1, high=3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_dimensions)).tolist()
    x_init = test_helpers.generate_random_test_quaternions(tensor_shape)

    self.assert_jacobian_is_correct_fn(quaternion.normalize, [x_init])

  @parameterized.parameters(
      ((4,), (4,)),
      ((None, 4), (None, 4)),
  )
  def test_multiply_exception_not_raised(self, *shapes):
    """Tests that the shape exceptions are not raised."""
    self.assert_exception_is_not_raised(quaternion.multiply, shapes)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (3,), (4,)),
      ("must have exactly 4 dimensions", (4,), (3,)),
  )
  def test_multiply_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.multiply, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_multiply_jacobian_preset(self):
    """Test the Jacobian of the multiply function."""
    x_1_init = test_helpers.generate_preset_test_quaternions()
    x_2_init = test_helpers.generate_preset_test_quaternions()

    self.assert_jacobian_is_correct_fn(quaternion.multiply,
                                       [x_1_init, x_2_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_multiply_jacobian_random(self):
    """Test the Jacobian of the multiply function."""
    tensor_dimensions = np.random.randint(low=1, high=3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_dimensions)).tolist()
    x_1_init = test_helpers.generate_random_test_quaternions(tensor_shape)
    x_2_init = test_helpers.generate_random_test_quaternions(tensor_shape)

    self.assert_jacobian_is_correct_fn(quaternion.multiply,
                                       [x_1_init, x_2_init])

  def test_normalized_random_initializer_raised(self):
    """Tests that the shape exceptions are raised."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()

    with self.subTest(name="dtype"):
      with self.assertRaisesRegexp(ValueError, "'dtype' must be tf.float32."):
        tf.compat.v1.get_variable(
            "test_variable",
            shape=tensor_shape + [4],
            dtype=tf.uint8,
            initializer=quaternion.normalized_random_uniform_initializer(),
            use_resource=False)

    with self.subTest(name="shape"):
      with self.assertRaisesRegexp(ValueError,
                                   "Last dimension of 'shape' must be 4."):
        tf.compat.v1.get_variable(
            "test_variable",
            shape=tensor_shape + [3],
            dtype=tf.float32,
            initializer=quaternion.normalized_random_uniform_initializer(),
            use_resource=False)

  def test_normalized_random_uniform_initializer_is_normalized(self):
    """Tests normalized_random_uniform_initializer outputs are normalized."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()

    variable = tf.compat.v1.get_variable(
        "test_variable",
        shape=tensor_shape + [4],
        dtype=tf.float32,
        initializer=quaternion.normalized_random_uniform_initializer(),
        use_resource=False)
    self.evaluate(tf.compat.v1.global_variables_initializer())
    value = self.evaluate(variable)
    norms = np.linalg.norm(value, axis=-1)
    ones = np.ones(tensor_shape)

    self.assertAllClose(norms, ones, rtol=1e-3)

  def test_normalized_random_uniform_is_normalized(self):
    """Tests that the normalized_random_uniform gives normalized quaternions."""
    tensor_size = np.random.randint(3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_size)).tolist()

    tensor = quaternion.normalized_random_uniform(tensor_shape)
    norms = tf.norm(tensor=tensor, axis=-1)
    ones = np.ones(tensor_shape)

    self.assertAllClose(norms, ones, rtol=1e-3)

  @parameterized.parameters(
      ((3,), (4,)),
      ((None, 3), (None, 4)),
  )
  def test_rotate_exception_not_raised(self, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_not_raised(quaternion.rotate, shape)

  @parameterized.parameters(
      ("must have exactly 3 dimensions", (2,), (4,)),
      ("must have exactly 4 dimensions", (3,), (2,)),
  )
  def test_rotate_exception_raised(self, error_msg, *shape):
    """Tests that the shape exceptions are raised."""
    self.assert_exception_is_raised(quaternion.rotate, error_msg, shape)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_preset(self):
    """Test the Jacobian of the rotate function."""
    x_matrix_init = test_helpers.generate_preset_test_quaternions()
    tensor_shape = x_matrix_init.shape[:-1] + (3,)
    x_point_init = np.random.uniform(size=tensor_shape)

    self.assert_jacobian_is_correct_fn(quaternion.rotate,
                                       [x_point_init, x_matrix_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_rotate_jacobian_random(self):
    """Test the Jacobian of the rotate function."""
    x_matrix_init = test_helpers.generate_random_test_quaternions()
    tensor_shape = x_matrix_init.shape[:-1] + (3,)
    x_point_init = np.random.uniform(size=tensor_shape)

    self.assert_jacobian_is_correct_fn(quaternion.rotate,
                                       [x_point_init, x_matrix_init])

  def test_rotate_random(self):
    """Tests the rotation using a quaternion vs a rotation matrix."""
    random_quaternion = test_helpers.generate_random_test_quaternions()
    tensor_shape = random_quaternion.shape[:-1]
    random_point = np.random.normal(size=tensor_shape + (3,))

    rotated_point_quaternion = quaternion.rotate(random_point,
                                                 random_quaternion)
    matrix = rotation_matrix_3d.from_quaternion(random_quaternion)
    rotated_point_matrix = rotation_matrix_3d.rotate(random_point, matrix)

    self.assertAllClose(
        rotated_point_matrix, rotated_point_quaternion, rtol=1e-3)

  @parameterized.parameters(
      ((td.QUAT_ID, td.QUAT_X_45), (np.pi / 4.0,)),
      ((td.QUAT_X_45, td.QUAT_ID), (np.pi / 4.0,)),
      ((td.QUAT_Y_90, td.QUAT_Y_180), (np.pi / 2.0,)),
      ((td.QUAT_X_180, td.QUAT_Z_180), (np.pi,)),
      ((td.QUAT_X_180, -1.0 * td.QUAT_Y_180), (np.pi,)),
      ((td.QUAT_X_180, td.QUAT_X_180), (0.0,)),
      ((td.QUAT_X_180, -1 * td.QUAT_X_180), (0.0,)),
      ((td.QUAT_X_90, td.QUAT_Y_90), (2 * np.pi / 3.0,)),
      ((np.array([0., 0., 0., 1]), np.array([0., 0., 0., 1])), (0.0,)),
  )
  def test_relative_angle(self, test_inputs, test_outputs):
    """Tests quaternion relative angle."""
    self.assert_output_is_correct(quaternion.relative_angle, test_inputs,
                                  test_outputs)

  @parameterized.parameters(
      ((4,), (4,)),
      ((None, 4), (None, 4)),
      ((None, None, 4), (None, None, 4)),
  )
  def test_relative_angle_not_raised(self, *shapes):
    """Tests that the shape exceptions of relative_angle are not raised."""
    self.assert_exception_is_not_raised(quaternion.relative_angle, shapes)

  @parameterized.parameters(
      ("must have exactly 4 dimensions", (3,), (4,)),
      ("must have exactly 4 dimensions", (4,), (3,)),
  )
  def test_relative_angle_raised(self, error_msg, *shape):
    """Tests that the shape exceptions of relative_angle are raised."""
    self.assert_exception_is_raised(quaternion.relative_angle, error_msg, shape)

  def test_valid_relative_angle_random(self):
    """Test the output is in valid range for relative_angle function."""
    tensor_dimensions = np.random.randint(low=1, high=3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_dimensions)).tolist()
    x_1_init = test_helpers.generate_random_test_quaternions(tensor_shape)
    x_2_init = test_helpers.generate_random_test_quaternions(tensor_shape)
    x_1 = tf.convert_to_tensor(value=x_1_init)
    x_2 = tf.convert_to_tensor(value=x_2_init)

    y = quaternion.relative_angle(x_1, x_2)

    self.assertAllGreaterEqual(y, 0.0)
    self.assertAllLessEqual(y, np.pi)

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_jacobian_relative_angle_random(self):
    """Test the Jacobian of the relative_angle function."""
    tensor_dimensions = np.random.randint(low=1, high=3)
    tensor_shape = np.random.randint(1, 10, size=(tensor_dimensions)).tolist()
    x_1_init = test_helpers.generate_random_test_quaternions(tensor_shape)
    x_2_init = test_helpers.generate_random_test_quaternions(tensor_shape)

    self.assert_jacobian_is_correct_fn(quaternion.relative_angle,
                                       [x_1_init, x_2_init])

  @flagsaver.flagsaver(tfg_add_asserts_to_graph=False)
  def test_jacobian_relative_angle_preset(self):
    """Test the Jacobian of the relative_angle function."""
    x_1_init = test_helpers.generate_preset_test_quaternions()
    x_2_init = test_helpers.generate_preset_test_quaternions()

    # relative angle is not smooth near <q1, q2> = 1, which occurs for
    # certain preset test quaternions.
    self.assert_jacobian_is_finite_fn(quaternion.relative_angle,
                                      [x_1_init, x_2_init])


if __name__ == "__main__":
  test_case.main()
